@@ -17,8 +17,8 @@
 //
 // Alternative licensing terms are available from the licensor.
 // For commercial licensing, see <https://www.artifex.com/> or contact
-// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
-// CA 94945, U.S.A., +1(415)492-9861, for further information.
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
 
 package com.artifex.mupdf.fitz;
 
@@ -83,7 +83,7 @@ abstract public class Device
 	abstract public void setDefaultColorSpaces(DefaultColorSpaces dcs);
 	abstract public void beginLayer(String name);
 	abstract public void endLayer();
-	abstract public void beginStructure(int standard, String raw, int uid);
+	abstract public void beginStructure(int standard, String raw, int idx);
 	abstract public void endStructure();
 	abstract public void beginMetatext(int meta, String text);
 	abstract public void endMetatext();
