@@ -172,7 +172,7 @@ img_lookup_metadata(fz_context *ctx, fz_document *doc_, const char *key, char *b
 }
 
 static fz_document *
-img_open_document(fz_context *ctx, fz_stream *file, fz_stream *accel, fz_archive *dir)
+img_open_document(fz_context *ctx, const fz_document_handler *handler, fz_stream *file, fz_stream *accel, fz_archive *dir, void *state)
 {
 	img_document *doc = fz_new_derived_document(ctx, img_document);
 
@@ -234,7 +234,7 @@ img_open_document(fz_context *ctx, fz_stream *file, fz_stream *accel, fz_archive
 }
 
 static int
-img_recognize_content(fz_context *ctx, fz_stream *stream, fz_archive *dir)
+img_recognize_content(fz_context *ctx, const fz_document_handler *handler, fz_stream *stream, fz_archive *dir, void **state, fz_document_recognize_state_free_fn **free_state)
 {
 	unsigned char data[8];
 	size_t n;
@@ -243,6 +243,11 @@ img_recognize_content(fz_context *ctx, fz_stream *stream, fz_archive *dir)
 	if (stream == NULL)
 		return 0;
 
+	if (state)
+		*state = NULL;
+	if (free_state)
+		*free_state = NULL;
+
 	n = fz_read(ctx, stream, data, 8);
 
 	if (n != 8)
