@@ -23,16 +23,133 @@
 #include "mupdf/fitz.h"
 
 #include <string.h>
+#ifndef _WIN32
+#include <unistd.h> /* For unlink */
+#endif
+#include <errno.h>
 
 enum
 {
 	FZ_DOCUMENT_HANDLER_MAX = 32
 };
 
+#ifndef PATH_MAX
+#define PATH_MAX 4096
+#endif
+
 #define DEFW (450)
 #define DEFH (600)
 #define DEFEM (12)
 
+static fz_output *
+fz_new_output_to_tempfile(fz_context *ctx, char **namep)
+{
+	fz_output *out = NULL;
+#ifdef _WIN32
+	char namebuf[L_tmpnam];
+	int attempts = 0;
+#else
+	char namebuf[] = "/tmp/fztmpXXXXXX";
+#endif
+
+
+	fz_var(out);
+
+#ifdef _WIN32
+	/* Windows has no mkstemp command, so we have to use the old-style
+	 * tmpnam based system, and retry in the case of races. */
+	do
+	{
+		if (tmpnam(namebuf) == NULL)
+			fz_throw(ctx, FZ_ERROR_SYSTEM, "tmpnam failed");
+		fz_try(ctx)
+			out = fz_new_output_with_path(ctx, namebuf, 0);
+		fz_catch(ctx)
+		{
+			/* We might hit a race condition and not be able to
+			 * open the file because someone beats us to it. We'd
+			 * be unbearably unlucky to hit this 10 times in a row. */
+			attempts++;
+			if (attempts >= 10)
+				fz_rethrow(ctx);
+			else
+				fz_ignore_error(ctx);
+		}
+	}
+	while (out == NULL);
+#else
+	{
+		FILE *file;
+		int fd = mkstemp(namebuf);
+
+		if (fd == -1)
+			fz_throw(ctx, FZ_ERROR_SYSTEM, "Cannot mkstemp: %s", strerror(errno));
+		file = fdopen(fd, "w");
+		if (file == NULL)
+			fz_throw(ctx, FZ_ERROR_SYSTEM, "Failed to open temporary file");
+		out = fz_new_output_with_file_ptr(ctx, file);
+	}
+#endif
+
+	if (namep)
+	{
+		fz_try(ctx)
+			*namep = fz_strdup(ctx, namebuf);
+		fz_catch(ctx)
+		{
+			fz_drop_output(ctx, out);
+			unlink(namebuf);
+			fz_rethrow(ctx);
+		}
+	}
+
+	return out;
+}
+
+static char *
+fz_new_tmpfile_from_stream(fz_context *ctx, fz_stream *stm)
+{
+	char *name;
+	fz_output *out = fz_new_output_to_tempfile(ctx, &name);
+
+	fz_try(ctx)
+	{
+		fz_write_stream(ctx, out, stm);
+		fz_close_output(ctx, out);
+	}
+	fz_always(ctx)
+		fz_drop_output(ctx, out);
+	fz_catch(ctx)
+	{
+		fz_free(ctx, name);
+		fz_rethrow(ctx);
+	}
+
+	return name;
+}
+
+static fz_stream *
+fz_file_backed_stream(fz_context *ctx, fz_stream *stream)
+{
+	const char *oname = fz_stream_filename(ctx, stream);
+	char *name;
+
+	/* If the file has a name, it's already a file-backed stream.*/
+	if (oname)
+		return stream;
+
+	/* Otherwise we need to make it one. */
+	name = fz_new_tmpfile_from_stream(ctx, stream);
+	fz_try(ctx)
+		stream = fz_open_file_autodelete(ctx, name);
+	fz_always(ctx)
+		fz_free(ctx, name);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
+
+	return stream;
+}
+
 struct fz_document_handler_context
 {
 	int refs;
@@ -55,9 +172,22 @@ fz_document_handler_context *fz_keep_document_handler_context(fz_context *ctx)
 
 void fz_drop_document_handler_context(fz_context *ctx)
 {
-	if (!ctx)
+	int i;
+
+	if (!ctx || !ctx->handler)
 		return;
 
+	for (i = 0; i < ctx->handler->count; i++)
+	{
+		if (ctx->handler->handler[i]->fin)
+		{
+			fz_try(ctx)
+				ctx->handler->handler[i]->fin(ctx, ctx->handler->handler[i]);
+			fz_catch(ctx)
+				fz_ignore_error(ctx);
+		}
+	}
+
 	if (fz_drop_imp(ctx, ctx->handler, &ctx->handler->refs))
 	{
 		fz_free(ctx, ctx->handler);
@@ -94,11 +224,20 @@ fz_recognize_document_stream_content(fz_context *ctx, fz_stream *stream, const c
 }
 
 const fz_document_handler *
-fz_recognize_document_stream_and_dir_content(fz_context *ctx, fz_stream *stream, fz_archive *dir, const char *magic)
+do_recognize_document_stream_and_dir_content(fz_context *ctx, fz_stream **streamp, fz_archive *dir, const char *magic, void **handler_state, fz_document_recognize_state_free_fn **handler_free_state)
 {
 	fz_document_handler_context *dc;
 	int i, best_score, best_i;
+	void *best_state = NULL;
+	fz_document_recognize_state_free_fn *best_free_state = NULL;
 	const char *ext;
+	int drop_stream = 0;
+	fz_stream *stream = *streamp;
+
+	if (handler_state)
+		*handler_state = NULL;
+	if (handler_free_state)
+		*handler_free_state = NULL;
 
 	dc = ctx->handler;
 	if (dc->count == 0)
@@ -113,94 +252,165 @@ fz_recognize_document_stream_and_dir_content(fz_context *ctx, fz_stream *stream,
 	best_score = 0;
 	best_i = -1;
 
-	if ((stream && stream->seek != NULL) || (stream == NULL && dir != NULL))
+	/* If we're handed a stream, check to see if any of our document handlers
+	 * need a file. If so, change the stream to be a file-backed one. */
+	if (stream)
 	{
+		int wants_file = 0;
 		for (i = 0; i < dc->count; i++)
-		{
-			int score = 0;
+			wants_file |= dc->handler[i]->wants_file;
 
-			if (dc->handler[i]->recognize_content)
-			{
-				if (stream)
-					fz_seek(ctx, stream, 0, SEEK_SET);
-				fz_try(ctx)
-				{
-					score = dc->handler[i]->recognize_content(ctx, stream, dir);
-				}
-				fz_catch(ctx)
-				{
-					/* in case of zip errors when recognizing EPUB/XPS/DOCX files */
-					fz_rethrow_unless(ctx, FZ_ERROR_FORMAT);
-					(void)fz_convert_error(ctx, NULL); /* ugly hack to silence the error message */
-					score = 0;
-				}
-			}
-			if (best_score < score)
-			{
-				best_score = score;
-				best_i = i;
-			}
+		/* Convert the stream into a file_backed stream. */
+		if (wants_file)
+		{
+			stream = fz_file_backed_stream(ctx, stream);
+			/* Either we need to pass this back to our caller, or we
+			 * need to drop it. */
+			drop_stream = 1;
 		}
-		if (stream)
-			fz_seek(ctx, stream, 0, SEEK_SET);
 	}
 
-	if (best_score < 100)
+	fz_try(ctx)
 	{
-		for (i = 0; i < dc->count; i++)
+		if ((stream && stream->seek != NULL) || (stream == NULL && dir != NULL))
 		{
-			int score = 0;
-			const char **entry;
-
-			if (dc->handler[i]->recognize)
-				score = dc->handler[i]->recognize(ctx, magic);
-
-			for (entry = &dc->handler[i]->mimetypes[0]; *entry; entry++)
-				if (!fz_strcasecmp(magic, *entry) && score < 100)
-				{
-					score = 100;
-					break;
-				}
-
-			if (ext)
+			for (i = 0; i < dc->count; i++)
 			{
-				for (entry = &dc->handler[i]->extensions[0]; *entry; entry++)
-					if (!fz_strcasecmp(ext, *entry) && score < 100)
+				void *state = NULL;
+				fz_document_recognize_state_free_fn *free_state = NULL;
+				int score = 0;
+
+				if (dc->handler[i]->recognize_content)
+				{
+					if (stream)
+						fz_seek(ctx, stream, 0, SEEK_SET);
+					fz_try(ctx)
+					{
+						score = dc->handler[i]->recognize_content(ctx, dc->handler[i], stream, dir, &state, &free_state);
+					}
+					fz_catch(ctx)
+					{
+						/* in case of zip errors when recognizing EPUB/XPS/DOCX files */
+						fz_rethrow_unless(ctx, FZ_ERROR_FORMAT);
+						(void)fz_convert_error(ctx, NULL); /* ugly hack to silence the error message */
+						score = 0;
+					}
+				}
+				if (best_score < score)
+				{
+					best_score = score;
+					best_i = i;
+					if (best_free_state)
+						best_free_state(ctx, best_state);
+					best_free_state = free_state;
+					best_state = state;
+				}
+				else if (free_state)
+					free_state(ctx, state);
+			}
+			if (stream)
+				fz_seek(ctx, stream, 0, SEEK_SET);
+		}
+
+		if (best_score < 100)
+		{
+			for (i = 0; i < dc->count; i++)
+			{
+				int score = 0;
+				const char **entry;
+
+				if (dc->handler[i]->recognize)
+					score = dc->handler[i]->recognize(ctx, dc->handler[i], magic);
+
+				for (entry = &dc->handler[i]->mimetypes[0]; *entry; entry++)
+					if (!fz_strcasecmp(magic, *entry) && score < 100)
 					{
 						score = 100;
 						break;
 					}
-			}
 
-			if (best_score < score)
-			{
-				best_score = score;
-				best_i = i;
+				if (ext)
+				{
+					for (entry = &dc->handler[i]->extensions[0]; *entry; entry++)
+						if (!fz_strcasecmp(ext, *entry) && score < 100)
+						{
+							score = 100;
+							break;
+						}
+				}
+
+				if (best_score < score)
+				{
+					best_score = score;
+					best_i = i;
+				}
 			}
 		}
 	}
+	fz_catch(ctx)
+	{
+		if (best_free_state)
+			best_free_state(ctx, best_state);
+		if (drop_stream)
+			fz_drop_stream(ctx, stream);
+		fz_rethrow(ctx);
+	}
 
 	if (best_i < 0)
+	{
+		if (drop_stream)
+			fz_drop_stream(ctx, stream);
 		return NULL;
+	}
+
+	/* Only if we found a handler, do we make our modified stream available to the
+	 * caller. */
+	*streamp = stream;
+
+	if (handler_state && handler_free_state)
+	{
+		*handler_state = best_state;
+		*handler_free_state = best_free_state;
+	}
+	else if (best_free_state)
+		best_free_state(ctx, best_state);
 
 	return dc->handler[best_i];
 }
 
-const fz_document_handler *fz_recognize_document_content(fz_context *ctx, const char *filename)
+const fz_document_handler *
+fz_recognize_document_stream_and_dir_content(fz_context *ctx, fz_stream *stream, fz_archive *dir, const char *magic)
+{
+	fz_stream *stm = stream;
+	const fz_document_handler *res;
+
+	res = do_recognize_document_stream_and_dir_content(ctx, &stm, dir, magic, NULL, NULL);
+
+	if (stm != stream)
+		fz_drop_stream(ctx, stm);
+
+	return res;
+}
+
+static const fz_document_handler *do_recognize_document_content(fz_context *ctx, const char *filename, void **handler_state, fz_document_recognize_state_free_fn **handler_free_state)
 {
 	fz_stream *stream = NULL;
 	const fz_document_handler *handler = NULL;
 	fz_archive *zip = NULL;
+	fz_stream *stm;
 
 	if (fz_is_directory(ctx, filename))
 		zip = fz_open_directory(ctx, filename);
 	else
 		stream  = fz_open_file(ctx, filename);
 
+	stm = stream;
 	fz_try(ctx)
-		handler = fz_recognize_document_stream_and_dir_content(ctx, stream, zip, filename);
+		handler = do_recognize_document_stream_and_dir_content(ctx, &stm, zip, filename, handler_state, handler_free_state);
 	fz_always(ctx)
 	{
+		if (stm != stream)
+			fz_drop_stream(ctx, stm);
 		fz_drop_stream(ctx, stream);
 		fz_drop_archive(ctx, zip);
 	}
@@ -210,6 +420,11 @@ const fz_document_handler *fz_recognize_document_content(fz_context *ctx, const
 	return handler;
 }
 
+const fz_document_handler *fz_recognize_document_content(fz_context* ctx, const char* filename)
+{
+	return do_recognize_document_content(ctx, filename, NULL, NULL);
+}
+
 const fz_document_handler *
 fz_recognize_document(fz_context *ctx, const char *magic)
 {
@@ -224,16 +439,34 @@ fz_document *
 fz_open_accelerated_document_with_stream_and_dir(fz_context *ctx, const char *magic, fz_stream *stream, fz_stream *accel, fz_archive *dir)
 {
 	const fz_document_handler *handler;
+	fz_stream *wrapped_stream = stream;
+	fz_document *ret;
+	void *state = NULL;
+	fz_document_recognize_state_free_fn *free_state = NULL;
 
 	if (stream == NULL && dir == NULL)
 		fz_throw(ctx, FZ_ERROR_ARGUMENT, "no document to open");
 	if (magic == NULL)
 		fz_throw(ctx, FZ_ERROR_ARGUMENT, "missing file type");
 
-	handler = fz_recognize_document_stream_and_dir_content(ctx, stream, dir, magic);
+	/* If this finds a handler, then this might wrap stream. If it does, we reuse the wrapped one in
+	 * the open call (hence avoiding us having to 'file-back' a stream twice), but we must free it. */
+	handler = do_recognize_document_stream_and_dir_content(ctx, &wrapped_stream, dir, magic, &state, &free_state);
 	if (!handler)
 		fz_throw(ctx, FZ_ERROR_UNSUPPORTED, "cannot find document handler for file type: '%s'", magic);
-	return handler->open(ctx, stream, accel, dir);
+	fz_try(ctx)
+		ret = handler->open(ctx, handler, wrapped_stream, accel, dir, state);
+	fz_always(ctx)
+	{
+		if (wrapped_stream != stream)
+			fz_drop_stream(ctx, wrapped_stream);
+		if (free_state && state)
+			free_state(ctx, state);
+	}
+	fz_catch(ctx)
+		fz_rethrow(ctx);
+
+	return ret;
 }
 
 fz_document *
@@ -272,23 +505,21 @@ fz_document *
 fz_open_accelerated_document(fz_context *ctx, const char *filename, const char *accel)
 {
 	const fz_document_handler *handler;
-	fz_stream *file;
+	fz_stream *file = NULL;
 	fz_stream *afile = NULL;
 	fz_document *doc = NULL;
-
-	fz_var(afile);
+	fz_archive *dir = NULL;
+	char dirname[PATH_MAX];
+	void *state = NULL;
+	fz_document_recognize_state_free_fn *free_state = NULL;
 
 	if (filename == NULL)
 		fz_throw(ctx, FZ_ERROR_ARGUMENT, "no document to open");
 
-	handler = fz_recognize_document_content(ctx, filename);
-	if (!handler)
-		fz_throw(ctx, FZ_ERROR_UNSUPPORTED, "cannot find document handler for file: %s", filename);
-
 	if (fz_is_directory(ctx, filename))
 	{
 		/* Cannot accelerate directories, currently. */
-		fz_archive *dir = fz_open_directory(ctx, filename);
+		dir = fz_open_directory(ctx, filename);
 
 		fz_try(ctx)
 			doc = fz_open_accelerated_document_with_stream_and_dir(ctx, filename, NULL, NULL, dir);
@@ -300,16 +531,31 @@ fz_open_accelerated_document(fz_context *ctx, const char *filename, const char *
 		return doc;
 	}
 
-	file = fz_open_file(ctx, filename);
+	handler = do_recognize_document_content(ctx, filename, &state, &free_state);
+	if (!handler)
+		fz_throw(ctx, FZ_ERROR_UNSUPPORTED, "cannot find document handler for file: %s", filename);
+
+	fz_var(afile);
+	fz_var(file);
 
 	fz_try(ctx)
 	{
+		file = fz_open_file(ctx, filename);
+
 		if (accel)
 			afile = fz_open_file(ctx, accel);
-		doc = handler->open(ctx, file, afile, NULL);
+		if (handler->wants_dir)
+		{
+			fz_dirname(dirname, filename, sizeof dirname);
+			dir = fz_open_directory(ctx, dirname);
+		}
+		doc = handler->open(ctx, handler, file, afile, dir, state);
 	}
 	fz_always(ctx)
 	{
+		if (free_state)
+			free_state(ctx, state);
+		fz_drop_archive(ctx, dir);
 		fz_drop_stream(ctx, afile);
 		fz_drop_stream(ctx, file);
 	}
