@@ -1098,6 +1098,23 @@ fz_dash_moveto(fz_context *ctx, struct sctx *s, float x, float y)
 	}
 }
 
+/*
+	Performs: a += (b-a) * i/n
+	allowing for FP inaccuracies that can cause a to "overrun" b.
+*/
+static float advance(float a, float b, float i, float n)
+{
+	float d = b - a;
+	float target = a + d * i/n;
+
+	if (d < 0 && target < b)
+		target = b;
+	else if (d > 0 && target > b)
+		target = b;
+
+	return target;
+}
+
 static void
 fz_dash_lineto(fz_context *ctx, struct sctx *s, float bx, float by, int from_bezier)
 {
@@ -1144,7 +1161,9 @@ fz_dash_lineto(fz_context *ctx, struct sctx *s, float bx, float by, int from_bez
 		}
 		ax = s->rect.x1;	/* d < 0, dx < 0 */
 a_moved_horizontally:	/* d and dx have the same sign */
-		ay += dy * d/dx;
+		assert((d > 0 && dx > 0) || (d < 0 && dx < 0));
+		assert(dx != 0);
+		ay = advance(ay, by, d, dx);
 		used = total * d/dx;
 		total -= used;
 		dx = bx - ax;
@@ -1176,7 +1195,9 @@ a_moved_horizontally:	/* d and dx have the same sign */
 		}
 		ay = s->rect.y1;	/* d < 0, dy < 0 */
 a_moved_vertically:	/* d and dy have the same sign */
-		ax += dx * d/dy;
+		assert((d > 0 && dy > 0) || (d < 0 && dy < 0));
+		assert(dy != 0);
+		ax = advance(ax, bx, d, dy);
 		d = total * d/dy;
 		total -= d;
 		used += d;
@@ -1224,7 +1245,11 @@ a_moved_vertically:	/* d and dy have the same sign */
 	}
 
 	/* Now if bx is off screen, bring it back */
-	if ((d = bx - s->rect.x0) < 0)
+	if (dx == 0)
+	{
+		/* Earlier stages can have moved a to be b, while leaving it completely off screen. */
+	}
+	else if ((d = bx - s->rect.x0) < 0)
 	{
 		old_bx = bx;
 		old_by = by;
@@ -1237,14 +1262,20 @@ a_moved_vertically:	/* d and dy have the same sign */
 		old_by = by;
 		bx = s->rect.x1;	/* d > 0, dx > 0 */
 b_moved_horizontally:	/* d and dx have the same sign */
-		by -= dy * d/dx;
+		assert((d > 0 && dx > 0) || (d < 0 && dx < 0));
+		assert(dx != 0);
+		by = advance(by, ay, d, dx);
 		tail = total * d/dx;
 		total -= tail;
 		dx = bx - ax;
 		dy = by - ay;
 	}
 	/* Then vertically... */
-	if ((d = by - s->rect.y0) < 0)
+	if (dy == 0)
+	{
+		/* Earlier stages can have moved a to be b, while leaving it completely off screen. */
+	}
+	else if ((d = by - s->rect.y0) < 0)
 	{
 		old_bx = bx;
 		old_by = by;
@@ -1258,7 +1289,9 @@ b_moved_horizontally:	/* d and dx have the same sign */
 		old_by = by;
 		by = s->rect.y1;	/* d > 0, dy > 0 */
 b_moved_vertically:	/* d and dy have the same sign */
-		bx -= dx * d/dy;
+		assert((d > 0 && dy > 0) || (d < 0 && dy < 0));
+		assert(dy != 0);
+		bx = advance(bx, ax, d, dy);
 		t = total * d/dy;
 		tail += t;
 		total -= t;
