@@ -1,3 +1,25 @@
+// Copyright (C) 2004-2021 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
+
 #include "mupdf/fitz.h"
 
 #include <string.h>
@@ -5,9 +27,7 @@
 
 /* TODO: check if this works with 16bpp images */
 
-typedef struct fz_predict_s fz_predict;
-
-struct fz_predict_s
+typedef struct
 {
 	fz_stream *chain;
 
@@ -24,7 +44,7 @@ struct fz_predict_s
 	unsigned char *rp, *wp;
 
 	unsigned char buffer[4096];
-};
+} fz_predict;
 
 static inline int getcomponent(unsigned char *line, int x, int bpc)
 {
@@ -222,7 +242,6 @@ close_predict(fz_context *ctx, void *state_)
 	fz_free(ctx, state);
 }
 
-/* Default values: predictor = 1, columns = 1, colors = 1, bpc = 8 */
 fz_stream *
 fz_open_predict(fz_context *ctx, fz_stream *chain, int predictor, int columns, int colors, int bpc)
 {
@@ -238,11 +257,11 @@ fz_open_predict(fz_context *ctx, fz_stream *chain, int predictor, int columns, i
 		bpc = 8;
 
 	if (bpc != 1 && bpc != 2 && bpc != 4 && bpc != 8 && bpc != 16)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "invalid number of bits per component: %d", bpc);
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "invalid number of bits per component: %d", bpc);
 	if (colors > FZ_MAX_COLORS)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "too many color components (%d > %d)", colors, FZ_MAX_COLORS);
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "too many color components (%d > %d)", colors, FZ_MAX_COLORS);
 	if (columns >= INT_MAX / (bpc * colors))
-		fz_throw(ctx, FZ_ERROR_GENERIC, "too many columns lead to an integer overflow (%d)", columns);
+		fz_throw(ctx, FZ_ERROR_LIMIT, "too many columns lead to an integer overflow (%d)", columns);
 
 	if (predictor != 1 && predictor != 2 &&
 			predictor != 10 && predictor != 11 &&
@@ -264,9 +283,9 @@ fz_open_predict(fz_context *ctx, fz_stream *chain, int predictor, int columns, i
 		state->stride = (state->bpc * state->colors * state->columns + 7) / 8;
 		state->bpp = (state->bpc * state->colors + 7) / 8;
 
-		state->in = fz_malloc(ctx, state->stride + 1);
-		state->out = fz_malloc(ctx, state->stride);
-		state->ref = fz_malloc(ctx, state->stride);
+		state->in = Memento_label(fz_malloc(ctx, state->stride + 1), "predict_in");
+		state->out = Memento_label(fz_malloc(ctx, state->stride), "predict_out");
+		state->ref = Memento_label(fz_malloc(ctx, state->stride), "predict_ref");
 		state->rp = state->out;
 		state->wp = state->out;
 
