@@ -17,8 +17,8 @@
 //
 // Alternative licensing terms are available from the licensor.
 // For commercial licensing, see <https://www.artifex.com/> or contact
-// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
-// CA 94945, U.S.A., +1(415)492-9861, for further information.
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
 
 #include "mupdf/fitz.h"
 
@@ -190,7 +190,7 @@ int fz_packed_path_size(const fz_path *path)
 }
 
 size_t
-fz_pack_path(fz_context *ctx, uint8_t *pack_, size_t max, const fz_path *path)
+fz_pack_path(fz_context *ctx, uint8_t *pack_, const fz_path *path)
 {
 	uint8_t *ptr;
 	size_t size;
@@ -201,9 +201,6 @@ fz_pack_path(fz_context *ctx, uint8_t *pack_, size_t max, const fz_path *path)
 		fz_packed_path *out = (fz_packed_path *)pack_;
 		size = sizeof(fz_packed_path) + sizeof(float) * pack->coord_len + sizeof(uint8_t) * pack->cmd_len;
 
-		if (size > max)
-			fz_throw(ctx, FZ_ERROR_GENERIC, "Can't pack a path that small!");
-
 		if (out)
 		{
 			out->refs = 1;
@@ -218,13 +215,10 @@ fz_pack_path(fz_context *ctx, uint8_t *pack_, size_t max, const fz_path *path)
 	size = sizeof(fz_packed_path) + sizeof(float) * path->coord_len + sizeof(uint8_t) * path->cmd_len;
 
 	/* If the path can't be packed flat, then pack it open */
-	if (path->cmd_len > 255 || path->coord_len > 255 || size > max)
+	if (path->cmd_len > 255 || path->coord_len > 255)
 	{
 		fz_path *pack = (fz_path *)pack_;
 
-		if (sizeof(fz_path) > max)
-			fz_throw(ctx, FZ_ERROR_GENERIC, "Can't pack a path that small!");
-
 		if (pack != NULL)
 		{
 			pack->refs = 1;
@@ -276,7 +270,7 @@ static void
 push_cmd(fz_context *ctx, fz_path *path, int cmd)
 {
 	if (path->refs != 1)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot modify shared paths");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "cannot modify shared paths");
 
 	if (path->cmd_len + 1 >= path->cmd_cap)
 	{
@@ -333,7 +327,7 @@ void
 fz_moveto(fz_context *ctx, fz_path *path, float x, float y)
 {
 	if (path->packed)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "Cannot modify a packed path");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "Cannot modify a packed path");
 
 	if (path->cmd_len > 0 && LAST_CMD(path) == FZ_MOVETO)
 	{
@@ -358,7 +352,7 @@ fz_lineto(fz_context *ctx, fz_path *path, float x, float y)
 	float x0, y0;
 
 	if (path->packed)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "Cannot modify a packed path");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "Cannot modify a packed path");
 
 	x0 = path->current.x;
 	y0 = path->current.y;
@@ -408,7 +402,7 @@ fz_curveto(fz_context *ctx, fz_path *path,
 	float x0, y0;
 
 	if (path->packed)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "Cannot modify a packed path");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "Cannot modify a packed path");
 
 	x0 = path->current.x;
 	y0 = path->current.y;
@@ -465,7 +459,7 @@ fz_quadto(fz_context *ctx, fz_path *path,
 	float x0, y0;
 
 	if (path->packed)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "Cannot modify a packed path");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "Cannot modify a packed path");
 
 	x0 = path->current.x;
 	y0 = path->current.y;
@@ -497,7 +491,7 @@ fz_curvetov(fz_context *ctx, fz_path *path, float x2, float y2, float x3, float
 	float x0, y0;
 
 	if (path->packed)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "Cannot modify a packed path");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "Cannot modify a packed path");
 
 	x0 = path->current.x;
 	y0 = path->current.y;
@@ -534,7 +528,7 @@ fz_curvetoy(fz_context *ctx, fz_path *path, float x1, float y1, float x3, float
 	float x0, y0;
 
 	if (path->packed)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "Cannot modify a packed path");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "Cannot modify a packed path");
 
 	x0 = path->current.x;
 	y0 = path->current.y;
@@ -566,7 +560,7 @@ fz_closepath(fz_context *ctx, fz_path *path)
 	uint8_t rep;
 
 	if (path->packed)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "Cannot modify a packed path");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "Cannot modify a packed path");
 
 	if (path->cmd_len == 0)
 	{
@@ -632,7 +626,7 @@ void
 fz_rectto(fz_context *ctx, fz_path *path, float x1, float y1, float x2, float y2)
 {
 	if (path->packed)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "Cannot modify a packed path");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "Cannot modify a packed path");
 
 	if (path->cmd_len > 0 && LAST_CMD(path) == FZ_MOVETO)
 	{
@@ -1019,7 +1013,7 @@ fz_transform_path(fz_context *ctx, fz_path *path, fz_matrix ctm)
 	fz_point p, p1, p2, p3, q, s;
 
 	if (path->packed)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "Cannot transform a packed path");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "Cannot transform a packed path");
 
 	if (ctm.b == 0 && ctm.c == 0)
 	{
@@ -1377,7 +1371,7 @@ fz_transform_path(fz_context *ctx, fz_path *path, fz_matrix ctm)
 void fz_trim_path(fz_context *ctx, fz_path *path)
 {
 	if (path->packed)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "Can't trim a packed path");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "Can't trim a packed path");
 	if (path->cmd_cap > path->cmd_len)
 	{
 		path->cmds = fz_realloc_array(ctx, path->cmds, path->cmd_len, unsigned char);
@@ -1611,7 +1605,7 @@ fz_clone_path(fz_context *ctx, fz_path *path)
 				}
 			}
 		default:
-			fz_throw(ctx, FZ_ERROR_GENERIC, "Unknown packing method found in path");
+			assert(!"Unknown packing method found in path");
 		}
 	}
 	fz_catch(ctx)
