@@ -748,7 +748,13 @@ do_subset(fz_context *ctx, cff_t *cff, fz_buffer **buffer, usage_list_t *keep_li
 	}
 
 	/* So we need 'required' bytes of space for the strings themselves */
-	offset_size = offsize_for_offset(required);
+	/* Do not forget to increment by one byte! This is because the
+	last entry in the offset table points to one byte beyond the end of
+	the required string data. Consider if the required string data occupies
+	255 bytes, then each offset for each of the required entries can be
+	represented by a single byte, but the last table entry would need to
+	point to offset 256, which cannot be represented by a single byte. */
+	offset_size = offsize_for_offset(required + 1);
 
 	required += 2 + 1 + (num_charstrings+1)*offset_size;
 
@@ -1258,15 +1264,15 @@ scan_charstrings(fz_context *ctx, cff_t *cff)
 	for (i = 0; i < num_charstrings; offset = end, i++)
 	{
 		end = index_get(ctx, &cff->charstrings_index, i+1);
-		if (i == 0)
-		{
-			/* Keep this. */
-		}
-		else if (gid < num_gids && i == gids[gid].num)
+		if (gid < num_gids && i == gids[gid].num)
 		{
 			/* Keep this */
 			gid++;
 		}
+		else if (i == 0)
+		{
+			/* Keep this. */
+		}
 		else
 		{
 			/* Drop this */
