@@ -505,9 +505,21 @@ subset_name_table(fz_context *ctx, ttf_t *ttf, fz_stream *stm)
 		new_len = 0;
 		for (i = 0; i < pl.len; i++)
 		{
-			uint32_t name_len = get16(pl.ptr[i] + 8);
-			uint8_t *name = d+off+get16(pl.ptr[i] + 10);
-			uint32_t offset = find_string_in_block(name, name_len, new_name_data, new_len);
+			uint32_t name_len, offset, name_off;
+			uint8_t *name;
+
+			if (t->len < (size_t) (pl.ptr[i] - t->data) + 8 + 2)
+				fz_throw(ctx, FZ_ERROR_FORMAT, "Truncated name length in name table");
+			name_len = get16(pl.ptr[i] + 8);
+
+			if (t->len < (size_t) (pl.ptr[i] - t->data) + 10 + 2)
+				fz_throw(ctx, FZ_ERROR_FORMAT, "Truncated name offset in name table");
+			name_off = off + get16(pl.ptr[i] + 10);
+			name = d + name_off;
+
+			if (t->len < name_off + name_len)
+				fz_throw(ctx, FZ_ERROR_FORMAT, "Truncated name in name table");
+			offset = find_string_in_block(name, name_len, new_name_data, new_len);
 			if (offset == UNFOUND)
 			{
 				if (name_data_size < new_len + name_len)
@@ -574,38 +586,53 @@ load_enc_tab4(fz_context *ctx, uint8_t *d, size_t data_size, uint32_t offset)
 	enc = fz_calloc(ctx, 1, sizeof(encoding_t) + sizeof(uint16_t) * (65536 - 256));
 	enc->max = 65536;
 
-	/* Run through the segments, counting how many are used. */
-	for (i = 0; i < seg_count; i++)
+	fz_try(ctx)
 	{
-		uint16_t seg_end = get16(d + offset + 14 + 2 * i);
-		uint16_t seg_start = get16(d + offset + 14 + 2 * seg_count + 2 + 2 * i);
-		uint16_t delta = get16(d + offset + 14 + 4 * seg_count + 2 + 2 * i);
-		uint32_t offset_ptr = offset + 14 + 6 * seg_count + 2 + 2 * i;
-		uint16_t offset = get16(d + offset_ptr);
-		uint16_t target;
-		uint32_t s;
-
-		if (seg_start >= enc->max || seg_end >= enc->max)
-			fz_throw(ctx, FZ_ERROR_FORMAT, "Malformed cmap4 table.");
-
-		for (s = seg_start; s <= seg_end; s++)
+		/* Run through the segments, counting how many are used. */
+		for (i = 0; i < seg_count; i++)
 		{
-			if (offset == 0)
-			{
-				target = delta + s;
-			}
-			else
-			{
-				/* Yes. This is very screwy. The offset is from the offset_ptr in use. */
-				target = get16(d + offset_ptr + offset + 2 * (s - seg_start));
-				if (target != 0)
-					target += delta;
-			}
+			uint16_t seg_end, seg_start, delta, target, inner_offset;
+			uint32_t offset_ptr, s;
 
-			if (target != 0)
-				enc->gid[s] = target;
+			if (data_size < offset + 14 + 6 * seg_count + 2 + 2 * i + 2)
+				fz_throw(ctx, FZ_ERROR_FORMAT, "cmap4 too small");
+
+			seg_end = get16(d + offset + 14 + 2 * i);
+			seg_start = get16(d + offset + 14 + 2 * seg_count + 2 + 2 * i);
+			delta = get16(d + offset + 14 + 4 * seg_count + 2 + 2 * i);
+			offset_ptr = offset + 14 + 6 * seg_count + 2 + 2 * i;
+			inner_offset = get16(d + offset_ptr);
+
+			if (seg_start >= enc->max || seg_end >= enc->max || seg_end < seg_start)
+				fz_throw(ctx, FZ_ERROR_FORMAT, "Malformed cmap4 table.");
+
+			for (s = seg_start; s <= seg_end; s++)
+			{
+				if (inner_offset == 0)
+				{
+					target = delta + s;
+				}
+				else
+				{
+					if (data_size < offset_ptr + inner_offset + 2 * (s - seg_start) + 2)
+						fz_throw(ctx, FZ_ERROR_FORMAT, "cmap4 too small");
+
+					/* Yes. This is very screwy. The inner_offset is from the offset_ptr in use. */
+					target = get16(d + offset_ptr + inner_offset + 2 * (s - seg_start));
+					if (target != 0)
+						target += delta;
+				}
+
+				if (target != 0)
+					enc->gid[s] = target;
+			}
 		}
 	}
+	fz_catch(ctx)
+	{
+		fz_free(ctx, enc);
+		fz_rethrow(ctx);
+	}
 
 	return enc;
 }
@@ -848,7 +875,13 @@ make_cmap(fz_context *ctx, ttf_t *ttf)
 		offset = 14 + segs * 2 * 3 + 2 + seg * 2;
 		put16(d + offset - segs * 2, 0); /* Delta - always 0 for now. */
 		put16(d + offset, entries - offset); /* offset */
-		put16(d + entries, ttf->is_otf ? enc->gid[i] : ttf->gid_renum[enc->gid[i]]); /* Insert an entry */
+
+		/* Insert an entry */
+		if (!ttf->is_otf && ttf->gid_renum && i < ttf->orig_num_glyphs && enc->gid[i] < ttf->orig_num_glyphs)
+			put16(d + entries, (ttf->is_otf || ttf->gid_renum == NULL) ? enc->gid[i] : ttf->gid_renum[enc->gid[i]]);
+		else
+			put16(d + entries, enc->gid[i]);
+
 		entries += 2;
 		for (i++; i < n; i++)
 		{
@@ -859,7 +892,10 @@ make_cmap(fz_context *ctx, ttf_t *ttf)
 				while (seg_end < i)
 				{
 					seg_end++;
-					put16(d + entries, ttf->is_otf ? enc->gid[seg_end] : ttf->gid_renum[enc->gid[seg_end]]);
+					if (!ttf->is_otf && ttf->gid_renum && seg_end < ttf->orig_num_glyphs && enc->gid[seg_end] < ttf->orig_num_glyphs)
+						put16(d + entries, ttf->gid_renum[enc->gid[seg_end]]);
+					else
+						put16(d + entries, enc->gid[seg_end]);
 					entries += 2;
 				}
 			}
@@ -970,14 +1006,17 @@ read_hhea(fz_context *ctx, ttf_t *ttf, fz_stream *stm)
 
 	ttf->orig_num_long_hor_metrics = get16(t->data+34);
 	if (ttf->orig_num_long_hor_metrics > ttf->orig_num_glyphs)
+	{
+		fz_drop_buffer(ctx, t);
 		fz_throw(ctx, FZ_ERROR_FORMAT, "Overlong hhea table");
+	}
 
 	add_table(ctx, ttf, TAG("hhea"), t);
 
 	/* Previously gids 0 to orig_num_long_hor_metrics-1 were described with
 	 * hor metrics, and the ones afterwards were fixed widths. Find where
 	 * that dividing line is in our new reduced set. */
-	if (ttf->encoding && !ttf->is_otf)
+	if (ttf->encoding && !ttf->is_otf && ttf->orig_num_long_hor_metrics > 0)
 	{
 		ttf->new_num_long_hor_metrics = 0;
 		for (i = ttf->orig_num_long_hor_metrics-1; i > 0; i--)
@@ -1033,15 +1072,15 @@ glyph_used(fz_context *ctx, ttf_t *ttf, fz_buffer *glyf, uint16_t i)
 	const uint8_t *data;
 	uint16_t flags;
 
-	if (ttf->gid_renum[i] != 0)
-		return;
-
-	if (i > ttf->orig_num_glyphs)
+	if (i >= ttf->orig_num_glyphs)
 	{
-		fz_warn(ctx, "TTF subsetting; gid > num_gids!");
+		fz_warn(ctx, "TTF subsetting; gid >= num_gids!");
 		return;
 	}
 
+	if (ttf->gid_renum[i] != 0)
+		return;
+
 	ttf->gid_renum[i] = 1;
 
 	/* If this glyf is composite, then we need to add any dependencies of it. */
@@ -1194,12 +1233,19 @@ read_glyf(fz_context *ctx, ttf_t *ttf, fz_stream *stm, int *gids, int num_gids)
 				glyph_used(ctx, ttf, t, enc->gid[i]);
 
 		/* Now convert from a usage table to a renumbering table. */
-		ttf->gid_renum[0] = 0;
-		j = 1;
-		for (i = 1; i < ttf->orig_num_glyphs; i++)
-			if (ttf->gid_renum[i])
-				ttf->gid_renum[i] = j++;
-		ttf->new_num_glyphs = j;
+		if (ttf->orig_num_glyphs > 0)
+		{
+			ttf->gid_renum[0] = 0;
+			j = 1;
+			for (i = 1; i < ttf->orig_num_glyphs; i++)
+				if (ttf->gid_renum[i])
+					ttf->gid_renum[i] = j++;
+			ttf->new_num_glyphs = j;
+		}
+		else
+		{
+			ttf->new_num_glyphs = 0;
+		}
 	}
 	else
 	{
@@ -1217,11 +1263,9 @@ read_glyf(fz_context *ctx, ttf_t *ttf, fz_stream *stm, int *gids, int num_gids)
 	for (i = 0; i < ttf->orig_num_glyphs; i++)
 	{
 		old_end = get_loca(ctx, ttf, i+1);
-		if (old_end > t->len)
+		if (old_end > t->len || old_end < old_start)
 			fz_throw(ctx, FZ_ERROR_FORMAT, "Bad loca value");
-		if (old_end == old_start)
-			continue;
-		if (i == 0 || ttf->gid_renum[i] != 0)
+		if ((old_end != old_start) && (i == 0 || ttf->gid_renum[i] != 0))
 		{
 			len = old_end - old_start;
 			memmove(t->data + new_start, t->data + old_start, len);
@@ -1458,7 +1502,13 @@ subset_post(fz_context *ctx, ttf_t *ttf, fz_stream *stm, int *gids, int num_gids
 		return;
 	}
 	d += 32; len -= 32;
-	len = subset_post2(ctx, ttf, d, len, gids, num_gids);
+	fz_try(ctx)
+		len = subset_post2(ctx, ttf, d, len, gids, num_gids);
+	fz_catch(ctx)
+	{
+		fz_drop_buffer(ctx, t);
+		fz_rethrow(ctx);
+	}
 
 	t->len = len;
 
