@@ -106,6 +106,7 @@ typedef struct marked_content_stack
 	struct marked_content_stack *next;
 	pdf_obj *tag;
 	pdf_obj *val;
+	int structure_pushed;
 } marked_content_stack;
 
 typedef struct begin_layer_stack
@@ -155,6 +156,10 @@ struct pdf_run_processor
 	/* Pending begin layers */
 	begin_layer_stack *begin_layer;
 	begin_layer_stack **next_begin_layer;
+
+	int mc_depth;
+	int nest_depth;
+	int nest_mark[1024];
 };
 
 static void
@@ -183,6 +188,12 @@ flush_begin_layer(fz_context *ctx, pdf_run_processor *proc)
 	while (proc->begin_layer)
 	{
 		s = proc->begin_layer;
+
+		if (proc->nest_depth == nelem(proc->nest_mark))
+			fz_throw(ctx, FZ_ERROR_LIMIT, "layer/clip nesting too deep");
+
+		proc->nest_mark[proc->nest_depth++] = ++proc->mc_depth;
+
 		fz_begin_layer(ctx, proc->dev, s->layer);
 		proc->begin_layer = s->next;
 		fz_free(ctx, s->layer);
@@ -191,6 +202,35 @@ flush_begin_layer(fz_context *ctx, pdf_run_processor *proc)
 	proc->next_begin_layer = &proc->begin_layer;
 }
 
+#define CLIP_MARK -1
+
+static void nest_layer_clip(fz_context *ctx, pdf_run_processor *proc)
+{
+	if (proc->nest_depth == nelem(proc->nest_mark))
+		fz_throw(ctx, FZ_ERROR_LIMIT, "layer/clip nesting too deep");
+	proc->nest_mark[proc->nest_depth++] = CLIP_MARK;
+}
+
+static void
+do_end_layer(fz_context *ctx, pdf_run_processor *proc)
+{
+	if (proc->nest_depth > 0 && proc->nest_mark[proc->nest_depth-1] == proc->mc_depth)
+	{
+		fz_end_layer(ctx, proc->dev);
+		proc->nest_depth--;
+	}
+	else
+	{
+		/* If EMC is unbalanced with q/Q, we will emit the end layer
+		 * device call before or after the Q operator instead of its true location
+		 */
+		 fz_warn(ctx, "invalid marked content and clip nesting");
+	}
+
+	if (proc->mc_depth > 0)
+		proc->mc_depth--;
+}
+
 typedef struct
 {
 	pdf_obj *softmask;
@@ -469,7 +509,22 @@ pdf_grestore(fz_context *ctx, pdf_run_processor *pr)
 	{
 		fz_try(ctx)
 		{
+			// End layer early (before Q) if unbalanced Q appears between BMC and EMC.
+			while (pr->nest_depth > 0 && pr->nest_mark[pr->nest_depth-1] != CLIP_MARK)
+			{
+				fz_end_layer(ctx, pr->dev);
+				pr->nest_depth--;
+			}
+
 			fz_pop_clip(ctx, pr->dev);
+			pr->nest_depth--;
+
+			// End layer late (after Q) if unbalanced EMC appears between q and Q.
+			while (pr->nest_depth > 0 && pr->nest_mark[pr->nest_depth-1] > pr->mc_depth)
+			{
+				fz_end_layer(ctx, pr->dev);
+				pr->nest_depth--;
+			}
 		}
 		fz_catch(ctx)
 		{
@@ -852,6 +907,7 @@ pdf_show_path(fz_context *ctx, pdf_run_processor *pr, int doclose, int dofill, i
 
 		if (pr->clip)
 		{
+			nest_layer_clip(ctx, pr);
 			gstate->clip_depth++;
 			fz_clip_path(ctx, pr->dev, path, pr->clip_even_odd, gstate->ctm, bbox);
 			pr->clip = 0;
@@ -1009,6 +1065,7 @@ pdf_flush_text(fz_context *ctx, pdf_run_processor *pr)
 
 		if (doclip)
 		{
+			nest_layer_clip(ctx, pr);
 			gstate->clip_depth++;
 			fz_clip_text(ctx, pr->dev, text, gstate->ctm, tb);
 		}
@@ -1518,7 +1575,7 @@ end_oc(fz_context *ctx, pdf_run_processor *proc, pdf_obj *val, pdf_cycle_list *c
 	if (obj)
 	{
 		flush_begin_layer(ctx, proc);
-		fz_end_layer(ctx, proc->dev);
+		do_end_layer(ctx, proc);
 		return;
 	}
 
@@ -1553,7 +1610,7 @@ end_layer(fz_context *ctx, pdf_run_processor *proc, pdf_obj *val)
 	pdf_obj *obj = pdf_dict_get(ctx, val, PDF_NAME(Title));
 	if (obj)
 	{
-		fz_end_layer(ctx, proc->dev);
+		do_end_layer(ctx, proc);
 	}
 }
 
@@ -1594,7 +1651,7 @@ pop_structure_to(fz_context *ctx, pdf_run_processor *proc, pdf_obj *common)
 	}
 #endif
 
-	while (pdf_objcmp(ctx, proc->mcid_sent, common))
+	while (proc->mcid_sent != NULL && pdf_objcmp(ctx, proc->mcid_sent, common))
 	{
 		pdf_obj *p = pdf_dict_get(ctx, proc->mcid_sent, PDF_NAME(P));
 		pdf_obj *tag = pdf_dict_get(ctx, proc->mcid_sent, PDF_NAME(S));
@@ -1696,7 +1753,7 @@ get_struct_index(fz_context *ctx, pdf_obj *send)
 	return -1;
 }
 
-static void
+static int
 send_begin_structure(fz_context *ctx, pdf_run_processor *proc, pdf_obj *mc_dict)
 {
 	pdf_obj *common = NULL;
@@ -1727,13 +1784,28 @@ send_begin_structure(fz_context *ctx, pdf_run_processor *proc, pdf_obj *mc_dict)
 		pdf_obj *slowptr = send;
 		int slow = 0;
 
+		/* Run up the ancestor stack, looking for the first child of mcid_sent.
+		 * That's the one we need to send next. */
 		while (1) {
 			pdf_obj *p = pdf_dict_get(ctx, send, PDF_NAME(P));
 
+			/* If we ever fail to find a dict, then do not step down lest
+			 * we can't get back later! */
+			if (!pdf_is_dict(ctx, send))
+			{
+				fz_warn(ctx, "Bad parent link in structure tree. Ignoring structure.");
+				proc->broken_struct_tree = 1;
+				return 0;
+			}
+			/* If p is the one we last sent, then we want to send 'send'
+			 * next. Exit the loop. */
 			if (!pdf_objcmp(ctx, p, proc->mcid_sent))
 				break;
+
+			/* We need to go at least one step further up the stack. */
 			send = p;
 
+			/* Check for a loop in the parent tree. */
 			slow ^= 1;
 			if (slow == 0)
 				slowptr = pdf_dict_get(ctx, slowptr, PDF_NAME(P));
@@ -1741,7 +1813,7 @@ send_begin_structure(fz_context *ctx, pdf_run_processor *proc, pdf_obj *mc_dict)
 			{
 				fz_warn(ctx, "Loop found in structure tree. Ignoring structure.");
 				proc->broken_struct_tree = 1;
-				return;
+				return 0;
 			}
 		}
 
@@ -1760,6 +1832,8 @@ send_begin_structure(fz_context *ctx, pdf_run_processor *proc, pdf_obj *mc_dict)
 #ifdef DEBUG_STRUCTURE
 	structure_dump(ctx, "on exit", proc->mcid_sent);
 #endif
+
+	return 1;
 }
 
 static void
@@ -1786,6 +1860,7 @@ push_marked_content(fz_context *ctx, pdf_run_processor *proc, const char *tagstr
 		mc->next = proc->marked_content;
 		mc->tag = tag;
 		mc->val = pdf_keep_obj(ctx, val);
+		mc->structure_pushed = 0;
 		proc->marked_content = mc;
 		drop_tag = 0;
 
@@ -1804,7 +1879,7 @@ push_marked_content(fz_context *ctx, pdf_run_processor *proc, const char *tagstr
 		if (mc_dict && !proc->broken_struct_tree)
 		{
 			fz_try(ctx)
-				send_begin_structure(ctx, proc, mc_dict);
+				mc->structure_pushed = send_begin_structure(ctx, proc, mc_dict);
 			fz_catch(ctx)
 			{
 				fz_report_error(ctx);
@@ -1846,6 +1921,7 @@ pop_marked_content(fz_context *ctx, pdf_run_processor *proc, int neat)
 	marked_content_stack *mc = proc->marked_content;
 	pdf_obj *val, *tag;
 	pdf_obj *mc_dict = NULL;
+	int pushed;
 
 	if (mc == NULL)
 		return;
@@ -1853,6 +1929,7 @@ pop_marked_content(fz_context *ctx, pdf_run_processor *proc, int neat)
 	proc->marked_content = mc->next;
 	tag = mc->tag;
 	val = mc->val;
+	pushed = mc->structure_pushed;
 	fz_free(ctx, mc);
 
 	/* If we're not interested in neatly closing any open layers etc
@@ -1887,7 +1964,7 @@ pop_marked_content(fz_context *ctx, pdf_run_processor *proc, int neat)
 		end_metatext(ctx, proc, val, mc_dict, PDF_NAME(ActualText));
 
 		/* Structure */
-		if (mc_dict && !proc->broken_struct_tree)
+		if (mc_dict && !proc->broken_struct_tree && pushed)
 		{
 			pdf_obj *p = pdf_dict_get(ctx, mc_dict, PDF_NAME(P));
 			pop_structure_to(ctx, proc, p);
@@ -2790,10 +2867,13 @@ pdf_close_run_processor(fz_context *ctx, pdf_processor *proc)
 	while (pr->gtop)
 		pdf_grestore(ctx, pr);
 
-	while (pr->gstate[0].clip_depth)
+	while (pr->nest_depth > 0)
 	{
-		fz_pop_clip(ctx, pr->dev);
-		pr->gstate[0].clip_depth--;
+		if (pr->nest_mark[pr->nest_depth-1] == CLIP_MARK)
+			fz_pop_clip(ctx, pr->dev);
+		else
+			fz_end_layer(ctx, pr->dev);
+		pr->nest_depth--;
 	}
 
 	pop_structure_to(ctx, pr, NULL);
