@@ -1,7 +1,7 @@
 //---------------------------------------------------------------------------------
 //
 //  Little Color Management System
-//  Copyright (c) 1998-2017 Marti Maria Saguer
+//  Copyright (c) 1998-2022 Marti Maria Saguer
 //
 // Permission is hereby granted, free of charge, to any person obtaining
 // a copy of this software and associated documentation files (the "Software"),
@@ -96,7 +96,7 @@ Error:
 
 // This function creates a profile based on White point, primaries and
 // transfer functions.
-cmsHPROFILE CMSEXPORT cmsCreateRGBProfileTHR(cmsContext ContextID,
+cmsHPROFILE CMSEXPORT cmsCreateRGBProfile(cmsContext ContextID,
                                           const cmsCIExyY* WhitePoint,
                                           const cmsCIExyYTRIPLE* Primaries,
                                           cmsToneCurve* const TransferFunction[3])
@@ -112,7 +112,7 @@ cmsHPROFILE CMSEXPORT cmsCreateRGBProfileTHR(cmsContext ContextID,
     if (!hICC)                          // can't allocate
         return NULL;
 
-    cmsSetProfileVersion(ContextID, hICC, 4.3);
+    cmsSetProfileVersion(ContextID, hICC, 4.4);
 
     cmsSetDeviceClass(ContextID, hICC,      cmsSigDisplayClass);
     cmsSetColorSpace(ContextID, hICC,       cmsSigRgbData);
@@ -212,17 +212,10 @@ Error:
     return NULL;
 }
 
-cmsHPROFILE CMSEXPORT cmsCreateRGBProfile(const cmsCIExyY* WhitePoint,
-                                          const cmsCIExyYTRIPLE* Primaries,
-                                          cmsToneCurve* const TransferFunction[3])
-{
-    return cmsCreateRGBProfileTHR(NULL, WhitePoint, Primaries, TransferFunction);
-}
-
 
 
 // This function creates a profile based on White point and transfer function.
-cmsHPROFILE CMSEXPORT cmsCreateGrayProfileTHR(cmsContext ContextID,
+cmsHPROFILE CMSEXPORT cmsCreateGrayProfile(cmsContext ContextID,
                                            const cmsCIExyY* WhitePoint,
                                            const cmsToneCurve* TransferFunction)
 {
@@ -233,7 +226,7 @@ cmsHPROFILE CMSEXPORT cmsCreateGrayProfileTHR(cmsContext ContextID,
     if (!hICC)                          // can't allocate
         return NULL;
 
-    cmsSetProfileVersion(ContextID, hICC, 4.3);
+    cmsSetProfileVersion(ContextID, hICC, 4.4);
 
     cmsSetDeviceClass(ContextID, hICC,      cmsSigDisplayClass);
     cmsSetColorSpace(ContextID, hICC,       cmsSigGrayData);
@@ -275,27 +268,22 @@ Error:
 
 
 
-cmsHPROFILE CMSEXPORT cmsCreateGrayProfile(const cmsCIExyY* WhitePoint,
-                                                    const cmsToneCurve* TransferFunction)
-{
-    return cmsCreateGrayProfileTHR(NULL, WhitePoint, TransferFunction);
-}
 
 // This is a devicelink operating in the target colorspace with as many transfer functions as components
 
-cmsHPROFILE CMSEXPORT cmsCreateLinearizationDeviceLinkTHR(cmsContext ContextID,
+cmsHPROFILE CMSEXPORT cmsCreateLinearizationDeviceLink(cmsContext ContextID,
                                                           cmsColorSpaceSignature ColorSpace,
                                                           cmsToneCurve* const TransferFunctions[])
 {
     cmsHPROFILE hICC;
     cmsPipeline* Pipeline;
-    cmsUInt32Number nChannels;
+    cmsInt32Number nChannels;
 
     hICC = cmsCreateProfilePlaceholder(ContextID);
     if (!hICC)
         return NULL;
 
-    cmsSetProfileVersion(ContextID, hICC, 4.3);
+    cmsSetProfileVersion(ContextID, hICC, 4.4);
 
     cmsSetDeviceClass(ContextID, hICC,      cmsSigLinkClass);
     cmsSetColorSpace(ContextID, hICC,       ColorSpace);
@@ -304,7 +292,7 @@ cmsHPROFILE CMSEXPORT cmsCreateLinearizationDeviceLinkTHR(cmsContext ContextID,
     cmsSetHeaderRenderingIntent(ContextID, hICC,  INTENT_PERCEPTUAL);
 
     // Set up channels
-    nChannels = cmsChannelsOf(ContextID, ColorSpace);
+    nChannels = cmsChannelsOfColorSpace(ContextID, ColorSpace);
 
     // Creates a Pipeline with prelinearization step only
     Pipeline = cmsPipelineAlloc(ContextID, nChannels, nChannels);
@@ -335,12 +323,6 @@ Error:
     return NULL;
 }
 
-cmsHPROFILE CMSEXPORT cmsCreateLinearizationDeviceLink(cmsColorSpaceSignature ColorSpace,
-                                                                 cmsToneCurve* const TransferFunctions[])
-{
-    return cmsCreateLinearizationDeviceLinkTHR(NULL, ColorSpace, TransferFunctions);
-}
-
 // Ink-limiting algorithm
 //
 //  Sum = C + M + Y + K
@@ -359,7 +341,7 @@ cmsHPROFILE CMSEXPORT cmsCreateLinearizationDeviceLink(cmsColorSpaceSignature Co
 //     K: Does not change
 
 static
-int InkLimitingSampler(cmsContext ContextID, register const cmsUInt16Number In[], register cmsUInt16Number Out[], register void* Cargo)
+int InkLimitingSampler(cmsContext ContextID, CMSREGISTER const cmsUInt16Number In[], CMSREGISTER cmsUInt16Number Out[], CMSREGISTER void* Cargo)
 {
     cmsFloat64Number InkLimit = *(cmsFloat64Number *) Cargo;
     cmsFloat64Number SumCMY, SumCMYK, Ratio;
@@ -367,7 +349,7 @@ int InkLimitingSampler(cmsContext ContextID, register const cmsUInt16Number In[]
 
     InkLimit = (InkLimit * 655.35);
 
-    SumCMY   = In[0]  + In[1] + In[2];
+    SumCMY   = (cmsFloat64Number) In[0]  + In[1] + In[2];
     SumCMYK  = SumCMY + In[3];
 
     if (SumCMYK > InkLimit) {
@@ -389,14 +371,14 @@ int InkLimitingSampler(cmsContext ContextID, register const cmsUInt16Number In[]
 
 // This is a devicelink operating in CMYK for ink-limiting
 
-cmsHPROFILE CMSEXPORT cmsCreateInkLimitingDeviceLinkTHR(cmsContext ContextID,
+cmsHPROFILE CMSEXPORT cmsCreateInkLimitingDeviceLink(cmsContext ContextID,
                                                      cmsColorSpaceSignature ColorSpace,
                                                      cmsFloat64Number Limit)
 {
     cmsHPROFILE hICC;
     cmsPipeline* LUT;
     cmsStage* CLUT;
-    cmsUInt32Number nChannels;
+    cmsInt32Number nChannels;
 
     if (ColorSpace != cmsSigCmykData) {
         cmsSignalError(ContextID, cmsERROR_COLORSPACE_CHECK, "InkLimiting: Only CMYK currently supported");
@@ -415,7 +397,7 @@ cmsHPROFILE CMSEXPORT cmsCreateInkLimitingDeviceLinkTHR(cmsContext ContextID,
     if (!hICC)                          // can't allocate
         return NULL;
 
-    cmsSetProfileVersion(ContextID, hICC, 4.3);
+    cmsSetProfileVersion(ContextID, hICC, 4.4);
 
     cmsSetDeviceClass(ContextID, hICC,      cmsSigLinkClass);
     cmsSetColorSpace(ContextID, hICC,       ColorSpace);
@@ -463,19 +445,14 @@ Error:
     return NULL;
 }
 
-cmsHPROFILE CMSEXPORT cmsCreateInkLimitingDeviceLink(cmsColorSpaceSignature ColorSpace, cmsFloat64Number Limit)
-{
-    return cmsCreateInkLimitingDeviceLinkTHR(NULL, ColorSpace, Limit);
-}
-
 
 // Creates a fake Lab identity.
-cmsHPROFILE CMSEXPORT cmsCreateLab2ProfileTHR(cmsContext ContextID, const cmsCIExyY* WhitePoint)
+cmsHPROFILE CMSEXPORT cmsCreateLab2Profile(cmsContext ContextID, const cmsCIExyY* WhitePoint)
 {
     cmsHPROFILE hProfile;
     cmsPipeline* LUT = NULL;
 
-    hProfile = cmsCreateRGBProfileTHR(ContextID, WhitePoint == NULL ? cmsD50_xyY(ContextID) : WhitePoint, NULL, NULL);
+    hProfile = cmsCreateRGBProfile(ContextID, WhitePoint == NULL ? cmsD50_xyY(ContextID) : WhitePoint, NULL, NULL);
     if (hProfile == NULL) return NULL;
 
     cmsSetProfileVersion(ContextID, hProfile, 2.1);
@@ -484,7 +461,7 @@ cmsHPROFILE CMSEXPORT cmsCreateLab2ProfileTHR(cmsContext ContextID, const cmsCIE
     cmsSetColorSpace(ContextID, hProfile,  cmsSigLabData);
     cmsSetPCS(ContextID, hProfile,         cmsSigLabData);
 
-    if (!SetTextTags(ContextID, hProfile, L"Lab identity built-in")) return NULL;
+    if (!SetTextTags(ContextID, hProfile, L"Lab identity built-in")) goto Error;
 
     // An identity LUT is all we need
     LUT = cmsPipelineAlloc(ContextID, 3, 3);
@@ -510,22 +487,17 @@ Error:
 }
 
 
-cmsHPROFILE CMSEXPORT cmsCreateLab2Profile(const cmsCIExyY* WhitePoint)
-{
-    return cmsCreateLab2ProfileTHR(NULL, WhitePoint);
-}
-
 
 // Creates a fake Lab V4 identity.
-cmsHPROFILE CMSEXPORT cmsCreateLab4ProfileTHR(cmsContext ContextID, const cmsCIExyY* WhitePoint)
+cmsHPROFILE CMSEXPORT cmsCreateLab4Profile(cmsContext ContextID, const cmsCIExyY* WhitePoint)
 {
     cmsHPROFILE hProfile;
     cmsPipeline* LUT = NULL;
 
-    hProfile = cmsCreateRGBProfileTHR(ContextID, WhitePoint == NULL ? cmsD50_xyY(ContextID) : WhitePoint, NULL, NULL);
+    hProfile = cmsCreateRGBProfile(ContextID, WhitePoint == NULL ? cmsD50_xyY(ContextID) : WhitePoint, NULL, NULL);
     if (hProfile == NULL) return NULL;
 
-    cmsSetProfileVersion(ContextID, hProfile, 4.3);
+    cmsSetProfileVersion(ContextID, hProfile, 4.4);
 
     cmsSetDeviceClass(ContextID, hProfile, cmsSigAbstractClass);
     cmsSetColorSpace(ContextID, hProfile,  cmsSigLabData);
@@ -556,22 +528,17 @@ Error:
     return NULL;
 }
 
-cmsHPROFILE CMSEXPORT cmsCreateLab4Profile(const cmsCIExyY* WhitePoint)
-{
-    return cmsCreateLab4ProfileTHR(NULL, WhitePoint);
-}
-
 
 // Creates a fake XYZ identity
-cmsHPROFILE CMSEXPORT cmsCreateXYZProfileTHR(cmsContext ContextID)
+cmsHPROFILE CMSEXPORT cmsCreateXYZProfile(cmsContext ContextID)
 {
     cmsHPROFILE hProfile;
     cmsPipeline* LUT = NULL;
 
-    hProfile = cmsCreateRGBProfileTHR(ContextID, cmsD50_xyY(ContextID), NULL, NULL);
+    hProfile = cmsCreateRGBProfile(ContextID, cmsD50_xyY(ContextID), NULL, NULL);
     if (hProfile == NULL) return NULL;
 
-    cmsSetProfileVersion(ContextID, hProfile, 4.3);
+    cmsSetProfileVersion(ContextID, hProfile, 4.4);
 
     cmsSetDeviceClass(ContextID, hProfile, cmsSigAbstractClass);
     cmsSetColorSpace(ContextID, hProfile,  cmsSigXYZData);
@@ -603,26 +570,21 @@ Error:
 }
 
 
-cmsHPROFILE CMSEXPORT cmsCreateXYZProfile(void)
-{
-    return cmsCreateXYZProfileTHR(NULL);
-}
-
 
 //sRGB Curves are defined by:
 //
-//If  RsRGB,GsRGB, BsRGB < 0.04045
+//If  R'sRGB,G'sRGB, B'sRGB < 0.04045
 //
-//    R =  RsRGB / 12.92
-//    G =  GsRGB / 12.92
-//    B =  BsRGB / 12.92
+//    R =  R'sRGB / 12.92
+//    G =  G'sRGB / 12.92
+//    B =  B'sRGB / 12.92
 //
 //
-//else if  RsRGB,GsRGB, BsRGB >= 0.04045
+//else if  R'sRGB,G'sRGB, B'sRGB >= 0.04045
 //
-//    R = ((RsRGB + 0.055) / 1.055)^2.4
-//    G = ((GsRGB + 0.055) / 1.055)^2.4
-//    B = ((BsRGB + 0.055) / 1.055)^2.4
+//    R = ((R'sRGB + 0.055) / 1.055)^2.4
+//    G = ((G'sRGB + 0.055) / 1.055)^2.4
+//    B = ((B'sRGB + 0.055) / 1.055)^2.4
 
 static
 cmsToneCurve* Build_sRGBGamma(cmsContext ContextID)
@@ -639,7 +601,7 @@ cmsToneCurve* Build_sRGBGamma(cmsContext ContextID)
 }
 
 // Create the ICC virtual profile for sRGB space
-cmsHPROFILE CMSEXPORT cmsCreate_sRGBProfileTHR(cmsContext ContextID)
+cmsHPROFILE CMSEXPORT cmsCreate_sRGBProfile(cmsContext ContextID)
 {
        cmsCIExyY       D65 = { 0.3127, 0.3290, 1.0 };
        cmsCIExyYTRIPLE Rec709Primaries = {
@@ -654,7 +616,7 @@ cmsHPROFILE CMSEXPORT cmsCreate_sRGBProfileTHR(cmsContext ContextID)
        Gamma22[0] = Gamma22[1] = Gamma22[2] = Build_sRGBGamma(ContextID);
        if (Gamma22[0] == NULL) return NULL;
 
-       hsRGB = cmsCreateRGBProfileTHR(ContextID, &D65, &Rec709Primaries, Gamma22);
+       hsRGB = cmsCreateRGBProfile(ContextID, &D65, &Rec709Primaries, Gamma22);
        cmsFreeToneCurve(ContextID, Gamma22[0]);
        if (hsRGB == NULL) return NULL;
 
@@ -666,11 +628,6 @@ cmsHPROFILE CMSEXPORT cmsCreate_sRGBProfileTHR(cmsContext ContextID)
        return hsRGB;
 }
 
-cmsHPROFILE CMSEXPORT cmsCreate_sRGBProfile(void)
-{
-    return cmsCreate_sRGBProfileTHR(NULL);
-}
-
 
 
 typedef struct {
@@ -685,7 +642,7 @@ typedef struct {
 
 
 static
-int bchswSampler(cmsContext ContextID, register const cmsUInt16Number In[], register cmsUInt16Number Out[], register void* Cargo)
+int bchswSampler(cmsContext ContextID, CMSREGISTER const cmsUInt16Number In[], CMSREGISTER cmsUInt16Number Out[], CMSREGISTER void* Cargo)
 {
     cmsCIELab LabIn, LabOut;
     cmsCIELCh LChIn, LChOut;
@@ -724,14 +681,14 @@ int bchswSampler(cmsContext ContextID, register const cmsUInt16Number In[], regi
 // Creates an abstract profile operating in Lab space for Brightness,
 // contrast, Saturation and white point displacement
 
-cmsHPROFILE CMSEXPORT cmsCreateBCHSWabstractProfileTHR(cmsContext ContextID,
-                                                       cmsUInt32Number nLUTPoints,
-                                                       cmsFloat64Number Bright,
-                                                       cmsFloat64Number Contrast,
-                                                       cmsFloat64Number Hue,
-                                                       cmsFloat64Number Saturation,
-                                                       cmsUInt32Number TempSrc,
-                                                       cmsUInt32Number TempDest)
+cmsHPROFILE CMSEXPORT cmsCreateBCHSWabstractProfile(cmsContext ContextID,
+                                                    cmsUInt32Number nLUTPoints,
+                                                    cmsFloat64Number Bright,
+                                                    cmsFloat64Number Contrast,
+                                                    cmsFloat64Number Hue,
+                                                    cmsFloat64Number Saturation,
+                                                    cmsUInt32Number TempSrc,
+                                                    cmsUInt32Number TempDest)
 {
     cmsHPROFILE hICC;
     cmsPipeline* Pipeline;
@@ -810,21 +767,10 @@ Error:
 }
 
 
-CMSAPI cmsHPROFILE   CMSEXPORT cmsCreateBCHSWabstractProfile(cmsUInt32Number nLUTPoints,
-                                                             cmsFloat64Number Bright,
-                                                             cmsFloat64Number Contrast,
-                                                             cmsFloat64Number Hue,
-                                                             cmsFloat64Number Saturation,
-                                                             cmsUInt32Number TempSrc,
-                                                             cmsUInt32Number TempDest)
-{
-    return cmsCreateBCHSWabstractProfileTHR(NULL, nLUTPoints, Bright, Contrast, Hue, Saturation, TempSrc, TempDest);
-}
-
 
 // Creates a fake NULL profile. This profile return 1 channel as always 0.
 // Is useful only for gamut checking tricks
-cmsHPROFILE CMSEXPORT cmsCreateNULLProfileTHR(cmsContext ContextID)
+cmsHPROFILE CMSEXPORT cmsCreateNULLProfile(cmsContext ContextID)
 {
     cmsHPROFILE hProfile;
     cmsPipeline* LUT = NULL;
@@ -838,7 +784,7 @@ cmsHPROFILE CMSEXPORT cmsCreateNULLProfileTHR(cmsContext ContextID)
     if (!hProfile)                          // can't allocate
         return NULL;
 
-    cmsSetProfileVersion(ContextID, hProfile, 4.3);
+    cmsSetProfileVersion(ContextID, hProfile, 4.4);
 
     if (!SetTextTags(ContextID, hProfile, L"NULL profile built-in")) goto Error;
 
@@ -882,11 +828,6 @@ Error:
     return NULL;
 }
 
-cmsHPROFILE CMSEXPORT cmsCreateNULLProfile(void)
-{
-    return cmsCreateNULLProfileTHR(NULL);
-}
-
 
 static
 int IsPCS(cmsColorSpaceSignature ColorSpace)
@@ -947,6 +888,8 @@ cmsHPROFILE CreateNamedColorDevicelink(cmsContext ContextID, cmsHTRANSFORM xform
     cmsHPROFILE hICC = NULL;
     cmsUInt32Number i, nColors;
     cmsNAMEDCOLORLIST *nc2 = NULL, *Original = NULL;
+    cmsUInt32Number InputFormat, OutputFormat;
+    cmsFormatter16 FromInput, ToOutput;
 
     // Create an empty placeholder
     hICC = cmsCreateProfilePlaceholder(ContextID);
@@ -954,7 +897,7 @@ cmsHPROFILE CreateNamedColorDevicelink(cmsContext ContextID, cmsHTRANSFORM xform
 
     // Critical information
     cmsSetDeviceClass(ContextID, hICC, cmsSigNamedColorClass);
-    cmsSetColorSpace(ContextID, hICC, v ->ExitColorSpace);
+    cmsSetColorSpace(ContextID, hICC, v ->core->ExitColorSpace);
     cmsSetPCS(ContextID, hICC, cmsSigLabData);
 
     // Tag profile with information
@@ -968,12 +911,33 @@ cmsHPROFILE CreateNamedColorDevicelink(cmsContext ContextID, cmsHTRANSFORM xform
     if (nc2 == NULL) goto Error;
 
     // Colorant count now depends on the output space
-    nc2 ->ColorantCount = cmsPipelineOutputChannels(ContextID, v ->Lut);
+    nc2 ->ColorantCount = cmsPipelineOutputChannels(ContextID, v ->core->Lut);
 
     // Make sure we have proper formatters
-    cmsChangeBuffersFormat(ContextID, xform, TYPE_NAMED_COLOR_INDEX,
-        FLOAT_SH(0) | COLORSPACE_SH(_cmsLCMScolorSpace(ContextID, v ->ExitColorSpace))
-        | BYTES_SH(2) | CHANNELS_SH(cmsChannelsOf(ContextID, v ->ExitColorSpace)));
+    // We only can afford to change formatters if previous transform is at least 16 bits
+    if (!(v->core->dwOriginalFlags & cmsFLAGS_CAN_CHANGE_FORMATTER)) {
+        cmsSignalError(ContextID, cmsERROR_NOT_SUITABLE,
+                       "CreateNamedColorDevicelink needs transforms created with at least 16 bits of precision");
+        goto Error;
+    }
+    InputFormat = TYPE_NAMED_COLOR_INDEX;
+    OutputFormat = FLOAT_SH(0) |
+                   COLORSPACE_SH(_cmsLCMScolorSpace(ContextID, v ->core->ExitColorSpace)) |
+                   BYTES_SH(2) |
+                   CHANNELS_SH(cmsChannelsOfColorSpace(ContextID, v ->core->ExitColorSpace));
+    FromInput = _cmsGetFormatter(ContextID, InputFormat, cmsFormatterInput, CMS_PACK_FLAGS_16BITS).Fmt16;
+    ToOutput  = _cmsGetFormatter(ContextID, OutputFormat, cmsFormatterOutput, CMS_PACK_FLAGS_16BITS).Fmt16;
+
+    if (FromInput == NULL || ToOutput == NULL) {
+        cmsSignalError(ContextID, cmsERROR_UNKNOWN_EXTENSION, "Unsupported raster format");
+        goto Error;
+    }
+
+    v ->InputFormat  = InputFormat;
+    v ->OutputFormat = OutputFormat;
+    v ->FromInput    = FromInput;
+    v ->ToOutput     = ToOutput;
+    _cmsFindFormatter(v, InputFormat, OutputFormat, v->core->dwOriginalFlags);
 
     // Apply the transfor to colorants.
     for (i=0; i < nColors; i++) {
@@ -1061,8 +1025,9 @@ const cmsAllowedLUT* FindCombination(cmsContext ContextID, const cmsPipeline* Lu
 cmsHPROFILE CMSEXPORT cmsTransform2DeviceLink(cmsContext ContextID, cmsHTRANSFORM hTransform, cmsFloat64Number Version, cmsUInt32Number dwFlags)
 {
     cmsHPROFILE hProfile = NULL;
-    cmsUInt32Number FrmIn, FrmOut, ChansIn, ChansOut;
-    int ColorSpaceBitsIn, ColorSpaceBitsOut;
+	cmsUInt32Number FrmIn, FrmOut;
+	cmsInt32Number ChansIn, ChansOut;
+	int ColorSpaceBitsIn, ColorSpaceBitsOut;
     _cmsTRANSFORM* xform = (_cmsTRANSFORM*) hTransform;
     cmsPipeline* LUT = NULL;
     cmsStage* mpe;
@@ -1073,7 +1038,7 @@ cmsHPROFILE CMSEXPORT cmsTransform2DeviceLink(cmsContext ContextID, cmsHTRANSFOR
     _cmsAssert(hTransform != NULL);
 
     // Get the first mpe to check for named color
-    mpe = cmsPipelineGetPtrToFirstStage(ContextID, xform ->Lut);
+    mpe = cmsPipelineGetPtrToFirstStage(ContextID, xform ->core->Lut);
 
     // Check if is a named color transform
     if (mpe != NULL) {
@@ -1084,19 +1049,20 @@ cmsHPROFILE CMSEXPORT cmsTransform2DeviceLink(cmsContext ContextID, cmsHTRANSFOR
     }
 
     // First thing to do is to get a copy of the transformation
-    LUT = cmsPipelineDup(ContextID, xform ->Lut);
+    LUT = cmsPipelineDup(ContextID, xform ->core->Lut);
     if (LUT == NULL) return NULL;
 
     // Time to fix the Lab2/Lab4 issue.
-    if ((xform ->EntryColorSpace == cmsSigLabData) && (Version < 4.0)) {
+    if ((xform ->core->EntryColorSpace == cmsSigLabData) && (Version < 4.0)) {
 
         if (!cmsPipelineInsertStage(ContextID, LUT, cmsAT_BEGIN, _cmsStageAllocLabV2ToV4curves(ContextID)))
             goto Error;
     }
 
-    // On the output side too
-    if ((xform ->ExitColorSpace) == cmsSigLabData && (Version < 4.0)) {
+    // On the output side too. Note that due to V2/V4 PCS encoding on lab we cannot fix white misalignments
+    if ((xform ->core->ExitColorSpace) == cmsSigLabData && (Version < 4.0)) {
 
+        dwFlags |= cmsFLAGS_NOWHITEONWHITEFIXUP;
         if (!cmsPipelineInsertStage(ContextID, LUT, cmsAT_END, _cmsStageAllocLabV4ToV2(ContextID)))
             goto Error;
     }
@@ -1107,15 +1073,15 @@ cmsHPROFILE CMSEXPORT cmsTransform2DeviceLink(cmsContext ContextID, cmsHTRANSFOR
 
     cmsSetProfileVersion(ContextID, hProfile, Version);
 
-    FixColorSpaces(ContextID, hProfile, xform -> EntryColorSpace, xform -> ExitColorSpace, dwFlags);
+    FixColorSpaces(ContextID, hProfile, xform ->core->EntryColorSpace, xform ->core->ExitColorSpace, dwFlags);
 
     // Optimize the LUT and precalculate a devicelink
 
-    ChansIn  = cmsChannelsOf(ContextID, xform -> EntryColorSpace);
-    ChansOut = cmsChannelsOf(ContextID, xform -> ExitColorSpace);
+    ChansIn  = cmsChannelsOfColorSpace(ContextID, xform -> core->EntryColorSpace);
+    ChansOut = cmsChannelsOfColorSpace(ContextID, xform -> core->ExitColorSpace);
 
-    ColorSpaceBitsIn  = _cmsLCMScolorSpace(ContextID, xform -> EntryColorSpace);
-    ColorSpaceBitsOut = _cmsLCMScolorSpace(ContextID, xform -> ExitColorSpace);
+    ColorSpaceBitsIn  = _cmsLCMScolorSpace(ContextID, xform ->core->EntryColorSpace);
+    ColorSpaceBitsOut = _cmsLCMScolorSpace(ContextID, xform ->core->ExitColorSpace);
 
     FrmIn  = COLORSPACE_SH(ColorSpaceBitsIn) | CHANNELS_SH(ChansIn)|BYTES_SH(2);
     FrmOut = COLORSPACE_SH(ColorSpaceBitsOut) | CHANNELS_SH(ChansOut)|BYTES_SH(2);
@@ -1136,7 +1102,7 @@ cmsHPROFILE CMSEXPORT cmsTransform2DeviceLink(cmsContext ContextID, cmsHTRANSFOR
     if (AllowedLUT == NULL) {
 
         // Try to optimize
-        _cmsOptimizePipeline(ContextID, &LUT, xform ->RenderingIntent, &FrmIn, &FrmOut, &dwFlags);
+        _cmsOptimizePipeline(ContextID, &LUT, xform->core->RenderingIntent, &FrmIn, &FrmOut, &dwFlags);
         AllowedLUT = FindCombination(ContextID, LUT, Version >= 4.0, DestinationTag);
 
     }
@@ -1148,7 +1114,7 @@ cmsHPROFILE CMSEXPORT cmsTransform2DeviceLink(cmsContext ContextID, cmsHTRANSFOR
         cmsStage* LastStage;
 
         dwFlags |= cmsFLAGS_FORCE_CLUT;
-        _cmsOptimizePipeline(ContextID, &LUT, xform ->RenderingIntent, &FrmIn, &FrmOut, &dwFlags);
+        _cmsOptimizePipeline(ContextID, &LUT, xform->core->RenderingIntent, &FrmIn, &FrmOut, &dwFlags);
 
         // Put identity curves if needed
         FirstStage = cmsPipelineGetPtrToFirstStage(ContextID, LUT);
@@ -1180,29 +1146,29 @@ cmsHPROFILE CMSEXPORT cmsTransform2DeviceLink(cmsContext ContextID, cmsHTRANSFOR
     if (!cmsWriteTag(ContextID, hProfile, DestinationTag, LUT)) goto Error;
 
 
-    if (xform -> InputColorant != NULL) {
-           if (!cmsWriteTag(ContextID, hProfile, cmsSigColorantTableTag, xform->InputColorant)) goto Error;
+    if (xform->core->InputColorant != NULL) {
+           if (!cmsWriteTag(ContextID, hProfile, cmsSigColorantTableTag, xform->core->InputColorant)) goto Error;
     }
 
-    if (xform -> OutputColorant != NULL) {
-           if (!cmsWriteTag(ContextID, hProfile, cmsSigColorantTableOutTag, xform->OutputColorant)) goto Error;
+    if (xform->core->OutputColorant != NULL) {
+           if (!cmsWriteTag(ContextID, hProfile, cmsSigColorantTableOutTag, xform->core->OutputColorant)) goto Error;
     }
 
-    if ((deviceClass == cmsSigLinkClass) && (xform ->Sequence != NULL)) {
-        if (!_cmsWriteProfileSequence(ContextID, hProfile, xform ->Sequence)) goto Error;
+    if ((deviceClass == cmsSigLinkClass) && (xform ->core->Sequence != NULL)) {
+        if (!_cmsWriteProfileSequence(ContextID, hProfile, xform ->core->Sequence)) goto Error;
     }
 
     // Set the white point
     if (deviceClass == cmsSigInputClass) {
-        if (!cmsWriteTag(ContextID, hProfile, cmsSigMediaWhitePointTag, &xform ->EntryWhitePoint)) goto Error;
+        if (!cmsWriteTag(ContextID, hProfile, cmsSigMediaWhitePointTag, &xform->core->EntryWhitePoint)) goto Error;
     }
     else {
-         if (!cmsWriteTag(ContextID, hProfile, cmsSigMediaWhitePointTag, &xform ->ExitWhitePoint)) goto Error;
+         if (!cmsWriteTag(ContextID, hProfile, cmsSigMediaWhitePointTag, &xform ->core->ExitWhitePoint)) goto Error;
     }
 
 
     // Per 7.2.15 in spec 4.3
-    cmsSetHeaderRenderingIntent(ContextID, hProfile, xform ->RenderingIntent);
+    cmsSetHeaderRenderingIntent(ContextID, hProfile, xform->core->RenderingIntent);
 
     cmsPipelineFree(ContextID, LUT);
     return hProfile;
